@@ -12,17 +12,22 @@ if PANDAS:
         return pd.DataFrame([x.as_record() for x in events])
 
 class Sentenai(API):
-    def __init__(self, host=None, port=None, protocol=None, check=True):
+    def __init__(self, host=None, port=None, check=True, auth=None):
         ## We do this so we can programmatically pass in host/port
         if host is None:
             host = 'localhost'
         if port is None:
-            port = 7280
-        if protocol is None:
+            if auth:
+                port = 443
+            else:
+                port = 7280
+        if auth:
+            protocol = 'https://'
+        else:
             protocol = 'http://'
 
         h = f"{protocol}{host}:{port}"
-        API.__init__(self, Credentials(h, ""))
+        API.__init__(self, Credentials(h, auth))
         if check and self.ping() > 0.5:
             print("warning: connection to this repository may be high latency or unstable.")
 
@@ -52,7 +57,7 @@ class Sentenai(API):
         representing nanoseconds since the origin `0`.
         """
         if origin == None:
-            r = self._put("db", name)
+            r = self._put("db", name, json={'origin': None})
         else:
             r = self._put("db", name, json={'origin': iso8601(origin)})
         if r.status_code != 201:
