@@ -15,6 +15,7 @@ from shapely.geometry import Point
 
 from queue import Queue
 from threading import Thread
+from concurrent.futures import ThreadPoolExecutor
 
 Update = namedtuple('Update', ['id', 'start', 'end', 'data'])
 
@@ -22,11 +23,11 @@ def index_data(args):
     db, node, index, v = args
     counter = 0
     while counter < 10:
+        data = cbor2.dumps(v)
         try:
             resp = db._post('nodes', node, 'types', index,
-                    json=cbor2.dumps(v), headers={'Content-Type': 'application/cbor'}, raw=True)
+                    json=data, headers={'Content-Type': 'application/cbor'}, raw=True)
         except:
-            print(v[:10])
             counter += 1
             sleep(.1)
         else:
@@ -292,52 +293,60 @@ class Database(API):
             dmap[cname] = []
 
         origin = self.origin
-        print(tmap)
-        for i, row in tqdm(df.iterrows(), total=len(df), unit='values', unit_scale=len(df.columns) - 1):
-            if origin is not None:
-                ts = (row['start'] - origin).delta
-            else:
-                ts = row['start']
-            try:
-                if 'end' in row and origin is not None:
-                    dur = (row['end'] - row['start']).delta
-                elif origin is not None:
-                    dur = (df['start'].iloc[i+1] - row['start']).delta
-                elif 'end' in row:
-                    dur = df['end'] - df['start']
+        res = []
+        with ThreadPoolExecutor(max_workers=64) as pool:
+            for i, row in tqdm(df.iterrows(), total=len(df), unit='values', unit_scale=len(df.columns) - 1):
+                if origin is not None:
+                    ts = (row['start'] - origin).delta
                 else:
-                    dur = (df['start'].iloc[i+1] - df['start']) // np.timedelta64(1, 'ns')
-            except IndexError:
-                dur = 1
-            else:
-                if dur <= 0: continue
-                for col, val in dict(row).items():
-                    if col == 'start':
-                        dmap[col].append((ts, dur))
-                    elif type(val) == float and math.isnan(val): # skip nans
-                        pass
-                    elif val is pd.NaT or val is None:
-                        pass
-                    elif tmap[col] == 'point3':
-                        dmap[col].append((ts, dur, (val.x, val.y, val.z)))
-                    elif tmap[col] == 'point':
-                        dmap[col].append((ts, dur, (val.x, val.y)))
-                    elif tmap[col] == 'date':
-                        dmap[col].append((ts, dur, val.isoformat()))
-                    elif tmap[col] == 'time':
-                        dmap[col].append((ts, dur, val.isoformat()))
-                    elif tmap[col] == 'datetime':
-                        dmap[col].append((ts, dur, iso8601(val)))
-                    elif tmap[col] == 'timedelta':
-                        dmap[col].append((ts, dur, val // np.timedelta64(1, 'ns')))
+                    ts = row['start'] // np.timedelta64(1, 'ns')
+                try:
+                    if 'end' in row and origin is not None:
+                        dur = (row['end'] - row['start']).delta
+                    elif origin is not None:
+                        dur = (df['start'].iloc[i+1] - row['start']).delta
+                    elif 'end' in row:
+                        dur = (row['end'] - row['start']) // np.timedelta64(1, 'ns')
                     else:
-                        dmap[col].append((ts, dur, val))
+                        dur = (df['start'].iloc[i+1] - row['start']) // np.timedelta64(1, 'ns')
+                except IndexError:
+                    dur = 1
+                else:
+                    if dur <= 0: continue
+                    for col, val in dict(row).items():
+                        if col == 'start':
+                            dmap[col].append((ts, dur))
+                        elif col == 'end':
+                            pass
+                        elif type(val) == float and math.isnan(val): # skip nans
+                            pass
+                        elif val is pd.NaT or val is None:
+                            pass
+                        elif tmap[col] == 'point3':
+                            dmap[col].append((ts, dur, (val.x, val.y, val.z)))
+                        elif tmap[col] == 'point':
+                            dmap[col].append((ts, dur, (val.x, val.y)))
+                        elif tmap[col] == 'date':
+                            dmap[col].append((ts, dur, val.isoformat()))
+                        elif tmap[col] == 'time':
+                            dmap[col].append((ts, dur, val.isoformat()))
+                        elif tmap[col] == 'datetime':
+                            dmap[col].append((ts, dur, iso8601(val)))
+                        elif tmap[col] == 'timedelta':
+                            dmap[col].append((ts, dur, val // np.timedelta64(1, 'ns')))
+                        else:
+                            dmap[col].append((ts, dur, val))
 
-            if len(dmap['start']) >= 4096:
-                with Pool(processes=16) as pool:
-                    pool.map(index_data, [(self, cmap[k], tmap[k], dmap[k]) for k, v in dmap.items()])
+                if len(dmap['start']) >= 4096:
+                    list(res) # force result
+                    res = pool.map(index_data, [(self, cmap[k], tmap[k], dmap[k]) for k, v in dmap.items()])
+                    for k, v in dmap.items():
+                        dmap[k] = []
+            if len(dmap['start']) > 0:
+                res = pool.map(index_data, [(self, cmap[k], tmap[k], dmap[k]) for k, v in dmap.items()])
                 for k, v in dmap.items():
                     dmap[k] = []
+                list(res) # force result
                 
 
 
@@ -379,6 +388,31 @@ class Stream(API):
         data = r.json()
         return iter(sorted(data.keys()))
 
+    def export(self, start=None, end=None, limit=None):
+        exp = API(self._credentials, "export")
+        o = iso8601(self._parent.origin)[:-1]
+        co = ["start", "end", "duration"] 
+        cols = [f"{self}/{x}" for x in iter(self)]
+        when = str(self)
+        params = {}
+        if limit is not None:
+            params['limit'] = limit
+        if start is not None:
+            params['start'] = iso8601(start)
+        if end is not None:
+            params['end'] = iso8601(end)
+
+        r = exp._post(json={'when': when, 'select': co+cols}, params=params)
+        return pd.DataFrame(
+                r.json(),
+                columns = co + list(self)
+                ).astype({
+                    'start': np.dtype('datetime64[ns]'),
+                    'end': np.dtype('datetime64[ns]'),
+                    'duration': np.dtype('timedelta64[ns]')
+                })
+
+
     @property
     def graph(self):
         return self._parent.graph.subtree(str(self))
@@ -391,7 +425,11 @@ class Stream(API):
     def type(self):
         r = self._get('types')
         if r.status_code == 200:
-            return r.json()[0]
+            ts = r.json()
+            if not ts:
+                return None
+            else:
+                return ts[0]
         else:
             return None
 
@@ -402,6 +440,8 @@ class Stream(API):
         r = self._get('types', self.type, 'range')
         if r.status_code == 200:
             e = r.json()
+            if e is None:
+                return None
             if self._parent.origin is None:
                 return (e['start'], e['end'])
             else:
